@@ -15,10 +15,9 @@ my $params = {
         'perl-Couchbase-Client/CouchbaseMock-0.5-SNAPSHOT.jar',
 
     #version numbers for libcouchbase and libvbucket
-    LIBVBUCKET_RELEASE => '1.8.0.2',
-    LIBCOUCHBASE_RELEASE => '1.0.0_75_g5622928',
+    LIBVBUCKET_RELEASE => '1.8.0.4',
+    LIBCOUCHBASE_RELEASE => '1.0.4',
 
-    #LIBISASL_RELEASE    => '1.0.0_3_g35e33e3',
     LIBEVENT_RELEASE    => '2.0.17-stable',
 };
 
