@@ -24,8 +24,9 @@ unset UMEM_DEBUG
 # We could have started it directly from the C code, but by using
 # a script it's a bit easier to test it manually ;)
 exec java \
+       -client \
        -jar tests/CouchbaseMock.jar \
         --nodes=10 \
-        --host=localhost \
+        --host=127.0.0.1 \
         --port=0 \
         "$@"
