@@ -26,14 +26,19 @@
 #include <sstream>
 #include <fstream>
 #include <string.h>
-#include <stdlib.h>
 #include <sys/stat.h>
 #include <cerrno>
+#include <cstdlib>
 #include <libcouchbase/couchbase.h>
-
 #include "configuration.h"
 #include "commandlineparser.h"
 
+#ifdef HAVE_LIBYAJL2
+#include <yajl/yajl_version.h>
+#include <yajl/yajl_parse.h>
+#include <yajl/yajl_tree.h>
+#endif
+
 using namespace std;
 
 enum cbc_command_t {
@@ -46,7 +51,8 @@ enum cbc_command_t {
     cbc_rm,
     cbc_stats,
     cbc_send,
-    cbc_version
+    cbc_version,
+    cbc_verify
 };
 
 extern "C" {
@@ -73,16 +79,16 @@ extern "C" {
     {
         if (error == LIBCOUCHBASE_SUCCESS) {
             cerr << "Stored \"";
-            cerr.write(static_cast<const char*>(key), nkey);
+            cerr.write(static_cast<const char *>(key), nkey);
             cerr << "\" CAS 0x" << hex << cas << endl;
         } else {
             cerr << "Failed to store \"";
-            cerr.write(static_cast<const char*>(key), nkey);
+            cerr.write(static_cast<const char *>(key), nkey);
             cerr << "\":" << endl
-                      << libcouchbase_strerror(instance, error) << endl;
+                 << libcouchbase_strerror(instance, error) << endl;
 
-            void *cookie = const_cast<void*>(libcouchbase_get_cookie(instance));
-            bool *e = static_cast<bool*>(cookie);
+            void *cookie = const_cast<void *>(libcouchbase_get_cookie(instance));
+            bool *e = static_cast<bool *>(cookie);
             *e = true;
         }
     }
@@ -94,15 +100,15 @@ extern "C" {
     {
         if (error == LIBCOUCHBASE_SUCCESS) {
             cerr << "Removed \"";
-            cerr.write(static_cast<const char*>(key), nkey);
+            cerr.write(static_cast<const char *>(key), nkey);
             cerr << "\"" << endl;
         } else {
             cerr << "Failed to remove \"";
-            cerr.write(static_cast<const char*>(key), nkey);
+            cerr.write(static_cast<const char *>(key), nkey);
             cerr << "\":" << endl
-                      << libcouchbase_strerror(instance, error) << endl;
-            void *cookie = const_cast<void*>(libcouchbase_get_cookie(instance));
-            bool *err = static_cast<bool*>(cookie);
+                 << libcouchbase_strerror(instance, error) << endl;
+            void *cookie = const_cast<void *>(libcouchbase_get_cookie(instance));
+            bool *err = static_cast<bool *>(cookie);
             *err = true;
         }
     }
@@ -116,60 +122,103 @@ extern "C" {
     {
         if (error == LIBCOUCHBASE_SUCCESS) {
             cerr << "\"";
-            cerr.write(static_cast<const char*>(key), nkey);
+            cerr.write(static_cast<const char *>(key), nkey);
             cerr << "\" Size " << nbytes << " Flags 0x" << std::hex
                  << flags << " CAS 0x" << cas << endl;
             cerr.flush();
-            cout.write(static_cast<const char*>(bytes), nbytes);
+            cout.write(static_cast<const char *>(bytes), nbytes);
             cout.flush();
         } else {
             cerr << "Failed to get \"";
-            cerr.write(static_cast<const char*>(key), nkey);
+            cerr.write(static_cast<const char *>(key), nkey);
             cerr << "\": " << libcouchbase_strerror(instance, error) << endl;
-            void *cookie = const_cast<void*>(libcouchbase_get_cookie(instance));
-            bool *err = static_cast<bool*>(cookie);
+            void *cookie = const_cast<void *>(libcouchbase_get_cookie(instance));
+            bool *err = static_cast<bool *>(cookie);
+            *err = true;
+        }
+    }
+
+    static void verify_callback(libcouchbase_t instance,
+                                const void *,
+                                libcouchbase_error_t error,
+                                const void *key, libcouchbase_size_t nkey,
+                                const void *bytes, libcouchbase_size_t nbytes,
+                                libcouchbase_uint32_t, libcouchbase_cas_t)
+    {
+        if (error == LIBCOUCHBASE_SUCCESS) {
+            char fnm[FILENAME_MAX];
+            memcpy(fnm, key, nkey);
+            fnm[nkey] = '\0';
+            struct stat st;
+            if (stat(fnm, &st) == -1) {
+                cerr << "Failed to look up: \"" << fnm << "\"" << endl;
+            } else if ((libcouchbase_size_t)st.st_size != nbytes) {
+                cerr << "Incorrect size for: \"" << fnm << "\"" << endl;
+            } else {
+                char *dta = new char[(libcouchbase_size_t)st.st_size];
+                if (dta == NULL) {
+                    cerr << "Failed to allocate memory to compare: \""
+                         << fnm << "\"" << endl;
+                } else {
+                    ifstream file(fnm, ios::binary);
+                    if (file.good() && file.read(dta, st.st_size) && file.good()) {
+                        if (memcmp(dta, bytes, nbytes) != 0) {
+                            cerr << "Content differ: \"" << fnm << "\"" << endl;
+                        }
+                    } else {
+                        cerr << "Failed to read \""  << fnm << "\"" << endl;
+                    }
+                    delete []dta;
+                }
+            }
+        } else {
+            cerr << "Failed to get \"";
+            cerr.write(static_cast<const char *>(key), nkey);
+            cerr << "\": " << libcouchbase_strerror(instance, error) << endl;
+            void *cookie = const_cast<void *>(libcouchbase_get_cookie(instance));
+            bool *err = static_cast<bool *>(cookie);
             *err = true;
         }
     }
 
     static void stat_callback(libcouchbase_t instance,
-                              const void*,
-                              const char* server_endpoint,
+                              const void *,
+                              const char *server_endpoint,
                               libcouchbase_error_t error,
-                              const void* key,
+                              const void *key,
                               libcouchbase_size_t nkey,
-                              const void* value,
+                              const void *value,
                               libcouchbase_size_t nvalue)
     {
         if (error == LIBCOUCHBASE_SUCCESS) {
             if (nkey > 0) {
                 cout << server_endpoint << "\t";
-                cout.write(static_cast<const char*>(key), nkey);
+                cout.write(static_cast<const char *>(key), nkey);
                 cout << "\t";
-                cout.write(static_cast<const char*>(value), nvalue);
+                cout.write(static_cast<const char *>(value), nvalue);
                 cout << endl;
             }
         } else {
             cerr << "Failure requesting stats:" << endl
                  << libcouchbase_strerror(instance, error) << endl;
 
-            void *cookie = const_cast<void*>(libcouchbase_get_cookie(instance));
-            bool *err = static_cast<bool*>(cookie);
+            void *cookie = const_cast<void *>(libcouchbase_get_cookie(instance));
+            bool *err = static_cast<bool *>(cookie);
             *err = true;
         }
     }
 
     static void flush_callback(libcouchbase_t instance,
-                               const void*,
-                               const char* server_endpoint,
+                               const void *,
+                               const char *server_endpoint,
                                libcouchbase_error_t error)
     {
         if (error != LIBCOUCHBASE_SUCCESS) {
             cerr << "Failed to flush node \"" << server_endpoint
                  << "\": " << libcouchbase_strerror(instance, error)
                  << endl;
-            void *cookie = const_cast<void*>(libcouchbase_get_cookie(instance));
-            bool *err = static_cast<bool*>(cookie);
+            void *cookie = const_cast<void *>(libcouchbase_get_cookie(instance));
+            bool *err = static_cast<bool *>(cookie);
             *err = true;
         }
     }
@@ -177,7 +226,8 @@ extern "C" {
     static void timings_callback(libcouchbase_t, const void *,
                                  libcouchbase_timeunit_t timeunit,
                                  libcouchbase_uint32_t min, libcouchbase_uint32_t max,
-                                 libcouchbase_uint32_t total, libcouchbase_uint32_t maxtotal) {
+                                 libcouchbase_uint32_t total, libcouchbase_uint32_t maxtotal)
+    {
         char buffer[1024];
         int offset = sprintf(buffer, "[%3u - %3u]", min, max);
         switch (timeunit) {
@@ -211,16 +261,40 @@ extern "C" {
     }
 }
 
-static bool cp(libcouchbase_t instance, list<string> &keys)
+static bool cp(libcouchbase_t instance, list<string> &keys, bool json)
 {
+    libcouchbase_size_t currsz = 0;
     for (list<string>::iterator ii = keys.begin(); ii != keys.end(); ++ii) {
-        std::string key = *ii;
+        string key = *ii;
         struct stat st;
         if (stat(key.c_str(), &st) == 0) {
-            char *bytes = new char[(libcouchbase_size_t)st.st_size];
+            char *bytes = new char[(libcouchbase_size_t)st.st_size + 1];
             if (bytes != NULL) {
                 ifstream file(key.c_str(), ios::binary);
                 if (file.good() && file.read(bytes, st.st_size) && file.good()) {
+                    bytes[st.st_size] = '\0';
+#ifdef HAVE_LIBYAJL2
+                    if (json) {
+                        yajl_val obj, id;
+                        const char *path[] = {"_id", NULL};
+                        if ((obj = yajl_tree_parse(bytes, NULL, 0)) == NULL) {
+                            cerr << "Failed to parse file \"" << key << "\" as JSON." << endl;
+                            delete []bytes;
+                            return false;
+                        }
+                        id = yajl_tree_get(obj, path, yajl_t_string);
+                        if (id == NULL) {
+                            cerr << "Failed to validate file \"" << key
+                                 << "\" as document (no '_id' attribute)." << endl;
+                            delete []bytes;
+                            return false;
+                        }
+                        key.assign(id->u.string);
+                        yajl_tree_free(obj);
+                    }
+#else
+                    (void)json;
+#endif
                     libcouchbase_store(instance,
                                        NULL,
                                        LIBCOUCHBASE_SET,
@@ -228,6 +302,14 @@ static bool cp(libcouchbase_t instance, list<string> &keys)
                                        bytes, (libcouchbase_size_t)st.st_size,
                                        0, 0, 0);
                     delete []bytes;
+                    currsz += (libcouchbase_size_t)st.st_size;
+
+                    // To avoid too much buffering flush at a regular
+                    // interval
+                    if (currsz > (2 * 1024 * 1024)) {
+                        libcouchbase_wait(instance);
+                        currsz = 0;
+                    }
                 } else {
                     cerr << "Failed to read file \"" << key << "\": "
                          << strerror(errno) << endl;
@@ -255,7 +337,7 @@ static bool rm(libcouchbase_t instance, list<string> &keys)
     }
 
     for (list<string>::iterator ii = keys.begin(); ii != keys.end(); ++ii) {
-        std::string key = *ii;
+        string key = *ii;
         libcouchbase_error_t err;
         err = libcouchbase_remove(instance, NULL, key.c_str(), key.length(), 0);
         if (err != LIBCOUCHBASE_SUCCESS) {
@@ -294,7 +376,7 @@ static bool cat(libcouchbase_t instance, list<string> &keys)
         cerr << "Failed to send requests:" << endl
              << libcouchbase_strerror(instance, err) << endl;
         return false;
-   }
+    }
 
     return true;
 }
@@ -311,7 +393,7 @@ static bool stats(libcouchbase_t instance, list<string> &keys)
         }
     } else {
         for (list<string>::iterator ii = keys.begin(); ii != keys.end(); ++ii) {
-            std::string key = *ii;
+            string key = *ii;
             libcouchbase_error_t err;
             err = libcouchbase_server_stats(instance, NULL, key.c_str(), key.length());
             if (err != LIBCOUCHBASE_SUCCESS) {
@@ -342,11 +424,12 @@ static bool flush(libcouchbase_t instance, list<string> &keys)
     return true;
 }
 
-static bool spool(string &data) {
+static bool spool(string &data)
+{
     stringstream ss;
     char buffer[1024];
     libcouchbase_size_t nr;
-    while ((nr = fread(buffer, 1, sizeof(buffer), stdin)) != (libcouchbase_size_t)-1) {
+    while ((nr = fread(buffer, 1, sizeof(buffer), stdin)) != (libcouchbase_size_t) - 1) {
         if (nr == 0) {
             break;
         }
@@ -395,6 +478,25 @@ static bool create(libcouchbase_t instance, list<string> &keys,
     return true;
 }
 
+static bool verify(libcouchbase_t instance, list<string> &keys)
+{
+    (void)libcouchbase_set_get_callback(instance, verify_callback);
+    return cat(instance, keys);
+}
+
+void loadKeys(list<string> &keys)
+{
+    char buffer[1024];
+    while (fgets(buffer, sizeof(buffer), stdin) != NULL) {
+        int idx = strlen(buffer);
+        while (idx > 0 && isspace(buffer[idx - 1])) {
+            --idx;
+        }
+        buffer[idx] = '\0';
+        keys.push_back(buffer);
+    }
+}
+
 extern bool receive(libcouchbase_t instance, list<string> &keys);
 extern bool send(libcouchbase_t instance, list<string> &keys);
 
@@ -415,6 +517,8 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
                                            "password for the rest port"));
     getopt.addOption(new CommandLineOption('T', "enable-timings", false,
                                            "Enable command timings"));
+    getopt.addOption(new CommandLineOption('t', "timeout", true,
+                                           "Specify timeout value"));
 
     libcouchbase_uint32_t flags = 0;
     libcouchbase_uint32_t exptime = 0;
@@ -428,12 +532,19 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
                                                "Fail if the object exists"));
     }
 
+    bool json = false;
+#ifdef HAVE_LIBYAJL2
+    if (cmd == cbc_cp) {
+        getopt.addOption(new CommandLineOption('j', "json", false,
+                                               "Treat value as JSON document (take key from '_id' attribute)"));
+    }
+#endif
     if (!getopt.parse(argc, argv)) {
         getopt.usage(argv[0]);
         exit(EXIT_FAILURE);
     }
 
-    vector<CommandLineOption*>::iterator iter;
+    vector<CommandLineOption *>::iterator iter;
     for (iter = getopt.options.begin(); iter != getopt.options.end(); ++iter) {
         if ((*iter)->found) {
             bool unknownOpt = true;
@@ -454,11 +565,19 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
                 config.setPassword((*iter)->argument);
                 break;
 
+            case 't' :
+                config.setTimeout((*iter)->argument);
+                break;
+
             case 'T' :
                 config.setTimingsEnabled(true);
                 break;
 
             case '?':
+                getopt.usage(argv[0]);
+                exit(EXIT_SUCCESS);
+                break;
+
             default:
                 if (cmd == cbc_create) {
                     unknownOpt = false;
@@ -475,6 +594,19 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
                     default:
                         unknownOpt = true;
                     }
+                    break;
+#ifdef HAVE_LIBYAJL2
+                } else if (cmd == cbc_cp) {
+                    unknownOpt = false;
+                    switch ((*iter)->shortopt) {
+                    case 'j':
+                        json = true;
+                        break;
+                    default:
+                        unknownOpt = true;
+                    }
+                    break;
+#endif
                 }
 
                 if (unknownOpt) {
@@ -490,7 +622,7 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
                                                   config.getPassword(),
                                                   config.getBucket(),
                                                   NULL);
-    if (instance == 0) {
+    if (instance == NULL) {
         cerr << "Failed to create couchbase instance" << endl;
         exit(EXIT_FAILURE);
     }
@@ -502,28 +634,39 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
     (void)libcouchbase_set_stat_callback(instance, stat_callback);
     (void)libcouchbase_set_storage_callback(instance, storage_callback);
 
+    if (config.getTimeout() != 0) {
+        libcouchbase_set_timeout(instance, config.getTimeout());
+    }
+
     libcouchbase_error_t ret = libcouchbase_connect(instance);
     if (ret != LIBCOUCHBASE_SUCCESS) {
         cerr << "Failed to connect libcouchbase instance to server:" << endl
-                  << "\t\"" << libcouchbase_strerror(instance, ret) << "\"" << endl;
+             << "\t\"" << libcouchbase_strerror(instance, ret) << "\"" << endl;
         exit(EXIT_FAILURE);
     }
     libcouchbase_wait(instance);
 
     bool error = false;
-    libcouchbase_set_cookie(instance, static_cast<void*>(&error));
+    libcouchbase_set_cookie(instance, static_cast<void *>(&error));
 
     if (config.isTimingsEnabled()) {
         libcouchbase_enable_timings(instance);
     }
 
+    list<string> keys;
+
     bool success;
     switch (cmd) {
     case cbc_cat:
         success = cat(instance, getopt.arguments);
         break;
     case cbc_cp:
-        success = cp(instance, getopt.arguments);
+        if (getopt.arguments.size() == 1 && getopt.arguments.front() == "-") {
+            loadKeys(keys);
+            success = cp(instance, keys, json);
+        } else {
+            success = cp(instance, getopt.arguments, json);
+        }
         break;
     case cbc_rm:
         success = rm(instance, getopt.arguments);
@@ -543,6 +686,14 @@ static void handleCommandLineOptions(enum cbc_command_t cmd, int argc, char **ar
     case cbc_create:
         success = create(instance, getopt.arguments, exptime, flags, add);
         break;
+    case cbc_verify:
+        if (getopt.arguments.size() == 1 && getopt.arguments.front() == "-") {
+            loadKeys(keys);
+            success = verify(instance, keys);
+        } else {
+            success = verify(instance, getopt.arguments);
+        }
+        break;
     default:
         cerr << "Not implemented" << endl;
         success = false;
@@ -594,6 +745,8 @@ static cbc_command_t getBuiltin(string name)
         return cbc_flush;
     } else if (name.find("cbc-version") != string::npos) {
         return cbc_version;
+    } else if (name.find("cbc-verify") != string::npos) {
+        return cbc_verify;
     }
 
     return cbc_illegal;
@@ -630,15 +783,26 @@ int main(int argc, char **argv)
                 cmd = cbc_flush;
             } else if (strcmp(argv[1], "version") == 0) {
                 cmd = cbc_version;
+            } else if (strcmp(argv[1], "verify") == 0) {
+                cmd = cbc_verify;
             }
         } else {
-            std::cerr << "Usage: cbc command [options]" << std::endl
-                      << "\tcommand may be: cat, cp, create, rm, stats, flush, version" << std::endl;
+            cerr << "Usage: cbc command [options]" << endl
+                 << "command may be:" << endl
+                 << "   cat        output keys to stdout" << endl
+                 << "   cp         store files to the cluster" << endl
+                 << "   create     store files with options" << endl
+                 << "   rm         remove keys" << endl
+                 << "   stats      show stats" << endl
+                 << "   flush      remove all keys from the cluster" << endl
+                 << "   version    show version" << endl
+                 << "   verify     verify content in cache with files" << endl
+                 << "Use 'cbc command --help' to show the options" << endl;
             exit(EXIT_FAILURE);
         }
 
         if (cmd == cbc_illegal) {
-            std::cerr << "Error: Unknown command \""<< argv[1] << "\"" << std::endl;
+            cerr << "Error: Unknown command \"" << argv[1] << "\"" << endl;
             exit(EXIT_FAILURE);
         }
         --argc;
@@ -649,6 +813,13 @@ int main(int argc, char **argv)
         cout << "cbc built from: " << PACKAGE_STRING << endl
              << "    using libcouchbase: " << libcouchbase_get_version(NULL)
              << endl;
+#ifdef HAVE_LIBYAJL2
+        int version = yajl_version();
+        cout << "    using libyajl: "
+             << version / 10000 << "."
+             << (version / 100) % 100 << "."
+             << version % 100 << endl;
+#endif
     } else {
         handleCommandLineOptions(cmd, argc, argv);
     }
